@@ -21,25 +21,46 @@
 #
 
 use strict;
+use Carp;
 
 require DBI;
 DBI->require_version(1.0201);
 
-use RPC::PlClient 0.2000;
+use RPC::PlClient 0.2000; # XXX change to 0.2017 once it's released
 
+{	package DBD::Proxy::RPC::PlClient;
+    	@DBD::Proxy::RPC::PlClient::ISA = qw(RPC::PlClient);
+	sub Call {
+	    my $self = shift;
+	    if ($self->{debug}) {
+		my ($rpcmeth, $obj, $method, @args) = @_;
+		local $^W; # silence undefs
+		Carp::carp("Server $rpcmeth $method(@args)");
+	    }
+	    return $self->SUPER::Call(@_);
+	}
+}
 
 
 package DBD::Proxy;
 
-use vars qw($VERSION $err $errstr $drh);
-
+use vars qw($VERSION $drh %ATTR);
 
 $VERSION = "0.2004";
 
-$err = 0;		# holds error code   for DBI::err
-$errstr = "";		# holds error string for DBI::errstr
 $drh = undef;		# holds driver handle once initialised
 
+%ATTR = (	# common to db & st, see also %ATTR in DBD::Proxy::db & ::st
+    'Warn'	=> 'local',
+    'Active'	=> 'local',
+    'Kids'	=> 'local',
+    'CachedKids' => 'local',
+    'PrintError' => 'local',
+    'RaiseError' => 'local',
+    'HandleError' => 'local',
+    'TraceLevel' => 'cached',
+    'CompatMode' => 'local',
+);
 
 sub driver ($$) {
     if (!$drh) {
@@ -50,20 +71,22 @@ sub driver ($$) {
 	$drh = DBI::_new_drh($class, {
 	    'Name' => 'Proxy',
 	    'Version' => $VERSION,
-	    'Err'    => \$DBD::Proxy::err,
-	    'Errstr' => \$DBD::Proxy::errstr,
 	    'Attribution' => 'DBD::Proxy by Jochen Wiedmann',
-	    });
+	});
+	$drh->STORE(CompatMode => 1); # disable DBI dispatcher attribute cache (for FETCH)
     }
     $drh;
 }
 
+sub CLONE {
+    undef $drh;
+}
+
 sub proxy_set_err {
   my ($h,$errmsg) = @_;
-  my ($err,$state) = 
-    ($errmsg =~ s/ \[err=(.*?),state=(.*?)\]//) ? ($1,$2) : (1,5 x ' ');
-
-  return DBI::set_err($h, $err, $errmsg, $state);
+  my ($err, $state) = ($errmsg =~ s/ \[err=(.*?),state=(.*?)\]//)
+	? ($1, $2) : (1, ' ' x 5);
+  return $h->set_err($err, $errmsg, $state);
 }
 
 package DBD::Proxy::dr; # ====== DRIVER ======
@@ -137,7 +160,7 @@ sub connect ($$;$$) {
 	}
     }
     # Create an RPC::PlClient object.
-    my($client, $msg) = eval { RPC::PlClient->new(%client_opts) };
+    my($client, $msg) = eval { DBD::Proxy::RPC::PlClient->new(%client_opts) };
 
     return DBD::Proxy::proxy_set_err($drh, "Cannot log in to DBI::ProxyServer: $@")
 	if $@; # Returns undef
@@ -189,8 +212,6 @@ sub connect ($$;$$) {
 }
 
 
-sub disconnect_all { }
-
 sub DESTROY { undef }
 
 
@@ -198,8 +219,20 @@ package DBD::Proxy::db; # ====== DATABASE ======
 
 $DBD::Proxy::db::imp_data_size = 0;
 
+# XXX probably many more methods need to be added here
+# in order to trigger our AUTOLOAD to redirect them to the server.
+# (Unless the sub is declared it's bypassed by perl method lookup.)
+# See notes in ToDo about method metadata
+# The question is whether to add all the methods in %DBI::DBI_methods
+# to the corresponding classes (::db, ::st etc)
+# Also need to consider methods that, if proxied, would change the server state
+# in a way that might not be visible on the client, ie begin_work -> AutoCommit.
+
 sub commit;
+sub connected;
 sub rollback;
+sub ping;
+
 
 use vars qw(%ATTR $AUTOLOAD);
 
@@ -211,67 +244,81 @@ use vars qw(%ATTR $AUTOLOAD);
 # Note: Attribute names starting with 'proxy_' always treated as 'inherited'.
 #
 %ATTR = (	# see also %ATTR in DBD::Proxy::st
-    'Warn' => 'local',
-    'Active' => 'local',
-    'Kids' => 'local',
-    'CachedKids' => 'local',
-    'Driver' => 'local',
-    'PrintError' => 'local',
-    'RaiseError' => 'local',
-    'RowCacheSize' => 'inherited',
-    'AutoCommit' => 'cached',
-    'Statement' => 'local',
+    %DBD::Proxy::ATTR,
+    RowCacheSize => 'inherited',
+    #AutoCommit => 'cached',
+    'FetchHashKeyName' => 'cached',
+    Statement => 'local',
+    Driver => 'local',
+    dbi_connect_closure => 'local',
+    Username => 'local',
 );
 
 sub AUTOLOAD {
     my $method = $AUTOLOAD;
     $method =~ s/(.*::(.*)):://;
-    # warn "AUTOLOAD of $method";
     my $class = $1;
     my $type = $2;
-    my %expand =
-	( 'method' => $method,
-	  'class' => $class,
-	  'type' => $type,
-	  'h' => "DBI::_::$type"
-	);
-    local $SIG{__DIE__} = 'DEFAULT';
-    my $method_code = UNIVERSAL::can($expand{'h'}, $method) ?
-	q/package ~class~;
-          sub ~method~ {
+    #warn "AUTOLOAD of $method (class=$class, type=$type)";
+    my %expand = (
+        'method' => $method,
+        'class' => $class,
+        'type' => $type,
+        'call' => "$method(\@_)",
+        # XXX was trying to be smart but was tripping up over the DBI's own
+        # smartness. Disabled, but left here in case there are issues.
+    #   'call' => (UNIVERSAL::can("DBI::_::$type", $method)) ? "$method(\@_)" : "func(\@_, '$method')",
+    );
+
+    my $method_code = q{
+        package ~class~;
+        sub ~method~ {
             my $h = shift;
-	    my @result = eval { $h->{'proxy_~type~h'}->~method~(@_) };
+            local $@;
+            my @result = wantarray
+                ? eval {        $h->{'proxy_~type~h'}->~call~ }
+                : eval { scalar $h->{'proxy_~type~h'}->~call~ };
             return DBD::Proxy::proxy_set_err($h, $@) if $@;
-            wantarray ? @result : $result[0];
-          }
-	 / :
-        q/package ~class~;
-	  sub ~method~ {
-	    my $h = shift;
-	    my @result = eval { $h->{'proxy_~type~h'}->func(@_, '~method~') };
-	    return DBD::Proxy::proxy_set_err($h, $@) if $@;
-	    wantarray ? @result : $result[0];
-          }
-         /;
+            return wantarray ? @result : $result[0];
+        }
+    };
     $method_code =~ s/\~(\w+)\~/$expand{$1}/eg;
-    eval $method_code;
-    die $@ if $@;
+    local $SIG{__DIE__} = 'DEFAULT';
+    my $err = do { local $@; eval $method_code.2; $@ };
+    die $err if $err;
     goto &$AUTOLOAD;
 }
 
 sub DESTROY {
-    # Just to avoid that DESTROY is autoloaded ...
+    my $dbh = shift;
+    local $@ if $@;	# protect $@
+    $dbh->disconnect if $dbh->SUPER::FETCH('Active');
 }
 
 sub disconnect ($) {
-    my($dbh) = @_;
-    # XXX this should call $rdbh->disconnect to get the right
-    # disconnect behaviour. It should not undef these values.
-    # A proxy_no_disconnect option could be added (like for finish)
-    # to let people trade safety for speed if they need to.
-    undef $dbh->{'proxy_dbh'};    # Bug in Perl 5.004; would prefer delete
-    undef $dbh->{'proxy_client'};
-    $this->SUPER::STORE('Active' => 0);
+    my ($dbh) = @_;
+
+    # Sadly the Proxy too-often disagrees with the backend database
+    # on the subject of 'Active'.  In the short term, I'd like the
+    # Proxy to ease up and let me decide when it's proper to go over
+    # the wire.  This ultimately applies to finish() as well.
+    #return unless $dbh->SUPER::FETCH('Active');
+
+    # Drop database connection at remote end
+    my $rdbh = $dbh->{'proxy_dbh'};
+    if ( $rdbh ) {
+        local $SIG{__DIE__} = 'DEFAULT';
+        local $@;
+	eval { $rdbh->disconnect() } ;
+        DBD::Proxy::proxy_set_err($dbh, $@) if $@;
+    }
+    
+    # Close TCP connect to remote
+    # XXX possibly best left till DESTROY? Add a config attribute to choose?
+    #$dbh->{proxy_client}->Disconnect(); # Disconnect method requires newer PlRPC module
+    $dbh->{proxy_client}->{socket} = undef; # hack
+
+    $dbh->SUPER::STORE('Active' => 0);
     1;
 }
 
@@ -280,16 +327,25 @@ sub STORE ($$$) {
     my($dbh, $attr, $val) = @_;
     my $type = $ATTR{$attr} || 'remote';
 
+    if ($attr eq 'TraceLevel') {
+	warn("TraceLevel $val");
+	my $pc = $dbh->{proxy_client} || die;
+	$pc->{logfile} ||= 1; # XXX hack
+	$pc->{debug} = ($val && $val >= 4);
+	$pc->Debug("$pc debug enabled") if $pc->{debug};
+    }
+
     if ($attr =~ /^proxy_/  ||  $type eq 'inherited') {
 	$dbh->{$attr} = $val;
 	return 1;
     }
 
-    if ($type eq 'remote'  ||  $type eq 'cached') {
+    if ($type eq 'remote' ||  $type eq 'cached') {
         local $SIG{__DIE__} = 'DEFAULT';
+	local $@;
 	my $result = eval { $dbh->{'proxy_dbh'}->STORE($attr => $val) };
 	return DBD::Proxy::proxy_set_err($dbh, $@) if $@; # returns undef
-	$dbh->{$attr} = $val if $type eq 'cached';
+	$dbh->SUPER::STORE($attr => $val) if $type eq 'cached';
 	return $result;
     }
     return $dbh->SUPER::STORE($attr => $val);
@@ -297,16 +353,18 @@ sub STORE ($$$) {
 
 sub FETCH ($$) {
     my($dbh, $attr) = @_;
+    # we only get here for cached attribute values if the handle is in CompatMode
+    # otherwise the DBI dispatcher handles the FETCH itself from the attribute cache.
     my $type = $ATTR{$attr} || 'remote';
 
-    if ($attr =~ /^proxy_/  ||  $type eq 'inherited'  ||
-	$type eq 'cached') {
+    if ($attr =~ /^proxy_/  ||  $type eq 'inherited'  || $type eq 'cached') {
 	return $dbh->{$attr};
     }
 
     return $dbh->SUPER::FETCH($attr) unless $type eq 'remote';
 
     local $SIG{__DIE__} = 'DEFAULT';
+    local $@;
     my $result = eval { $dbh->{'proxy_dbh'}->FETCH($attr) };
     return DBD::Proxy::proxy_set_err($dbh, $@) if $@;
     return $result;
@@ -326,6 +384,7 @@ sub prepare ($$;$) {
       $sth->{'proxy_attr_cache'} = {cache_filled => 0};
       my $rdbh = $dbh->{'proxy_dbh'};
       local $SIG{__DIE__} = 'DEFAULT';
+      local $@;
       my $rsth = eval { $rdbh->prepare($sth->{'Statement'}, $sth->{'proxy_attr'}, undef, $proto_ver) };
       return DBD::Proxy::proxy_set_err($sth, $@) if $@;
       return DBD::Proxy::proxy_set_err($sth, "Constructor didn't return a handle: $rsth")
@@ -337,6 +396,8 @@ sub prepare ($$;$) {
       $sth->{'proxy_sth'} = $rsth;
       # If statement is a positioned update we do not want any readahead.
       $sth->{'RowCacheSize'} = 1 if $stmt =~ /\bfor\s+update\b/i;
+    # Since resources are used by prepared remote handle, mark us active.
+    $sth->SUPER::STORE(Active => 1);
     }
     $sth;
 }
@@ -362,6 +423,7 @@ sub quote {
     # for example.
     # Jochen
     local $SIG{__DIE__} = 'DEFAULT';
+    local $@;
     my $result = eval { $dbh->{'proxy_dbh'}->quote(@_) };
     return DBD::Proxy::proxy_set_err($dbh, $@) if $@;
     return $result;
@@ -372,9 +434,10 @@ sub table_info {
     my $rdbh = $dbh->{'proxy_dbh'};
     #warn "table_info(@_)";
     local $SIG{__DIE__} = 'DEFAULT';
+    local $@;
     my($numFields, $names, $types, @rows) = eval { $rdbh->table_info(@_) };
     return DBD::Proxy::proxy_set_err($dbh, $@) if $@;
-    my $sth = DBI::_new_sth($dbh, {
+    my ($sth, $inner) = DBI::_new_sth($dbh, {
         'Statement' => "SHOW TABLES",
 	'proxy_params' => [],
 	'proxy_data' => \@rows,
@@ -388,7 +451,9 @@ sub table_info {
     	'proxy_cache_only' => 1,
     });
     $sth->SUPER::STORE('NUM_OF_FIELDS' => $numFields);
-    $sth->SUPER::STORE('Active' => 1);
+    $inner->{NAME} = $names;
+    $inner->{TYPE} = $types;
+    $sth->SUPER::STORE('Active' => 1); # already execute()'d
     $sth->{'proxy_rows'} = @rows;
     return $sth;
 }
@@ -403,6 +468,7 @@ sub tables {
 sub type_info_all {
     my $dbh = shift;
     local $SIG{__DIE__} = 'DEFAULT';
+    local $@;
     my $result = eval { $dbh->{'proxy_dbh'}->type_info_all(@_) };
     return DBD::Proxy::proxy_set_err($dbh, $@) if $@;
     return $result;
@@ -425,12 +491,7 @@ use vars qw(%ATTR);
 # Note: Attribute names starting with 'proxy_' always treated as 'inherited'.
 #
 %ATTR = (	# see also %ATTR in DBD::Proxy::db
-    'Warn' => 'local',
-    'Active' => 'local',
-    'Kids' => 'local',
-    'CachedKids' => 'local',
-    'PrintError' => 'local',
-    'RaiseError' => 'local',
+    %DBD::Proxy::ATTR,
     'Database' => 'local',
     'RowsInCache' => 'local',
     'RowCacheSize' => 'inherited',
@@ -459,6 +520,7 @@ sub execute ($@) {
     my ($numRows, @outData);
 
     local $SIG{__DIE__} = 'DEFAULT';
+    local $@;
     if ( $proto_ver > 1 ) {
       ($numRows, @outData) = eval { $rsth->execute($params, $proto_ver) };
       return DBD::Proxy::proxy_set_err($sth, $@) if $@;
@@ -539,6 +601,7 @@ sub fetch ($) {
 	}
 	my $num_rows = $sth->FETCH('RowCacheSize') || 20;
 	local $SIG{__DIE__} = 'DEFAULT';
+	local $@;
 	my @rows = eval { $rsth->fetch($num_rows) };
 	return DBD::Proxy::proxy_set_err($sth, $@) if $@;
 	unless (@rows == $num_rows) {
@@ -572,6 +635,7 @@ sub finish ($) {
 	: $sth->FETCH('Database')->{'proxy_no_finish'};
     unless ($no_finish) {
         local $SIG{__DIE__} = 'DEFAULT';
+	local $@;
 	my $result = eval { $rsth->finish() };
 	return DBD::Proxy::proxy_set_err($sth, $@) if $@;
 	return $result;
@@ -592,12 +656,13 @@ sub STORE ($$$) {
 	return 0;
     }
 
-    if ($type eq 'remote') {
+    if ($type eq 'remote' || $type eq 'cached') {
 	my $rsth = $sth->{'proxy_sth'}  or  return undef;
         local $SIG{__DIE__} = 'DEFAULT';
+	local $@;
 	my $result = eval { $rsth->STORE($attr => $val) };
 	return DBD::Proxy::proxy_set_err($sth, $@) if ($@);
-	return $result;
+	return $result if $type eq 'remote'; # else fall through to cache locally
     }
     return $sth->SUPER::STORE($attr => $val);
 }
@@ -625,6 +690,7 @@ sub FETCH ($$) {
     if ($type ne 'local') {
 	my $rsth = $sth->{'proxy_sth'}  or  return undef;
         local $SIG{__DIE__} = 'DEFAULT';
+	local $@;
 	my $result = eval { $rsth->FETCH($attr) };
 	return DBD::Proxy::proxy_set_err($sth, $@) if $@;
 	return $result;
@@ -645,7 +711,8 @@ sub bind_param ($$$@) {
 *bind_param_inout = \&bind_param;
 
 sub DESTROY {
-    # Just to avoid autoloading DESTROY ...
+    my $sth = shift;
+    $sth->finish if $sth->SUPER::FETCH('Active');
 }
 
 
@@ -695,7 +762,7 @@ I<Note that the connect_cached method is new and still experimental.>
 Before connecting to a remote database, you must ensure, that a Proxy
 server is running on the remote machine. There's no default port, so
 you have to ask your system administrator for the port number. See
-L<DBI::ProxyServer(3)> for details.
+L<DBI::ProxyServer> for details.
 
 Say, your Proxy server is running on machine "alpha", port 3334, and
 you'd like to connect to an ODBC database called "mydb" as user "joe"
@@ -712,7 +779,7 @@ You see, this is mainly the same. The DBD::Proxy module will create a
 connection to the Proxy server on "alpha" which in turn will connect
 to the ODBC database.
 
-Refer to the L<DBI(3)> documentation on the C<connect> method for a way
+Refer to the L<DBI> documentation on the C<connect> method for a way
 to automatically use DBD::Proxy without having to change your code.
 
 DBD::Proxy's DSN string has the format
@@ -762,7 +829,7 @@ by executing
     $cipherRef = $class->new(pack("H*", $key));
 
 and pass this object to the RPC::PlClient module when creating a
-client. See L<RPC::PlClient(3)>. Example:
+client. See L<RPC::PlClient>. Example:
 
     cipher=IDEA;key=97cd2375efa329aceef2098babdc9721
 
@@ -774,7 +841,7 @@ less secure than the usercipher/userkey secret and readable by anyone.
 The usercipher/userkey secret is B<your> private secret.
 
 Of course encryption requires an appropriately configured server. See
-<DBD::ProxyServer(3)/CONFIGURATION FILE>.
+<DBD::ProxyServer/CONFIGURATION FILE>.
 
 =item debug
 
@@ -821,7 +888,7 @@ and fast CGI applications.
 This attribute can be used to reduce network traffic: By default calls
 to $dbh->quote() are passed to the remote driver.  Of course this slows
 down things quite a lot, but is the safest default behaviour.
-  
+
 However, if you set the I<proxy_quote> attribute to the value 'C<local>'
 either in the database handle or in the statement handle, and the call
 to quote has only one parameter, then the local default DBI quote
@@ -831,6 +898,26 @@ method will be used (which will be faster but may be wrong).
 
 =head1 KNOWN ISSUES
 
+=head2 Unproxied method calls
+
+If a method isn't being proxied, try declaring a stub sub in the appropriate
+package (DBD::Proxy::db for a dbh method, and DBD::Proxy::st for an sth method).
+For example:
+
+    sub DBD::Proxy::db::selectall_arrayref;
+
+That will enable selectall_arrayref to be proxied.
+
+Currently many methods aren't explicitly proxied and so you get the DBI's
+default methods executed on the client.
+
+Some of those methods, like selectall_arrayref, may then call other methods
+that are proxied (selectall_arrayref calls fetchall_arrayref which calls fetch
+which is proxied). So things may appear to work but operate more slowly than
+the could.
+
+This may all change in a later version.
+
 =head2 Complex handle attributes
 
 Sometimes handles are having complex attributes like hash refs or
@@ -851,13 +938,13 @@ executing the above example in two steps:
 
 =over
 
-=item 1.)
- 
+=item 1
+
 The first step is fetching the value of the key "csv_tables" in the
 handle $dbh. The value returned is complex, a hash ref.
- 
-=item 2.)
- 
+
+=item 2
+
 The second step is storing some value (the right hand side of the
 assignment) as the key "passwd" in the hash ref from step 1.
 
@@ -901,6 +988,6 @@ is granted to Tim Bunce for distributing this as a part of the DBI.
 
 =head1 SEE ALSO
 
-L<DBI(3)>, L<RPC::PlClient(3)>, L<Storable(3)>
+L<DBI>, L<RPC::PlClient>, L<Storable>
 
 =cut
